/************************************************************************/
/* TaxSolve_NC_D400_2012.c - North Carolina 2012 NC-DC400 State Taxes.	*/
/* Copyright (C) 2012 - S.Jenkins					*/
/* 									*/
/* Compile:   gcc taxsolve_NC_D400_2012.c -o taxsolve_NC_D400_2012	*/
/* Run:  ./taxsolve_NC_D400_2012		  			*/
/*  Uses log from TaxSolve Federal 1040.				*/
/* 									*/
/* GNU Public License - GPL:						*/
/* This program is free software; you can redistribute it and/or	*/
/* modify it under the terms of the GNU General Public License as	*/
/* published by the Free Software Foundation; either version 2 of the	*/
/* License, or (at your option) any later version.			*/
/* 									*/
/* This program is distributed in the hope that it will be useful,	*/
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU	*/
/* General Public License for more details.				*/
/* 									*/
/* You should have received a copy of the GNU General Public License	*/
/* along with this program; if not, write to the Free Software		*/
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA		*/
/* 02111-1307 USA							*/
/* 									*/
/* 1-8-2013	http://opentaxsolver.sourceforge.com/			*/
/*			Lincoln Baxter (lab@lincolnbaxter.com)		*/
/*									*/
/************************************************************************/

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <ctype.h>

#include "taxsolve_routines.c"

float thisversion=10.0;

#define SINGLE                  1
#define MARRIED_FILLING_JOINTLY 2
#define MARRIED_FILLING_SEPARAT 3
#define HEAD_OF_HOUSEHOLD       4
#define WIDOW                   5
#define Yes 1
#define No  0

/* Modified by Lincoln Baxter for 2012 NC-Rate Schedule on page 29
 * of the D-400 instruction book.
 * */
double TaxRateFunction( double x, int status )
{
 switch (status)
  {
   case SINGLE:				/*Single*/
	if (x < 12750.0) return 0.06 * x; else
	if (x < 60000.0) return 765.0 + 0.07 * (x-12750.0);
	else return 4072.5 + 0.0775 * (x-60000.0); 
	break;
   case HEAD_OF_HOUSEHOLD:		/*Head of Household*/
	if (x < 17000.0) return 0.06 * x; else
	if (x < 80000.0) return 1020.0 + 0.07 * (x-17000.0);
	else return 5430.0 + 0.0775 * (x-80000.0);
	break;
   case MARRIED_FILLING_JOINTLY:	/*Married/joint*/
   case WIDOW:	/*Widow(er)*/
	if (x < 21250.0) return 0.06 * x; else
	if (x < 100000.0) return 1275.0 + 0.07 * (x-21250.0);
	else return 6787.5 + 0.0775 * (x-100000.0);
	break;
   case MARRIED_FILLING_SEPARAT:	/*Married/Separate*/
	if (x < 10625.0) return 0.06 * x; else
	if (x < 50000.0) return 637.5 + 0.07 * (x-10625.0);
	else return 3393.75 + 0.0775 * (x-50000.0);
	break;
  }
 return 0.0; 
}


struct FedReturnData
 {
  double fedline[MAX_LINES];
  int Itemized, Limited, Limited_L6;
  double Sched_A[MAX_LINES];
 };


void convert_slashes( char *fname )
{ /* Convert slashes in file name based on machine type. */
  char *ptr;
 #ifdef __MINGW32__
  char slash_sreach='/', slash_replace='\\';
 #else
  char slash_sreach='\\', slash_replace='/';
 #endif

  ptr = strchr( fname, slash_sreach );
  while (ptr)
   {
    ptr[0] = slash_replace;
    ptr = strchr( fname, slash_sreach );
   }
}


void ImportFederalReturnData( char *fedlogfile, struct FedReturnData *fed_data )
{
 FILE *infile;
 char fline[1000], word[1000];
 int linenum;

 convert_slashes( fedlogfile );
 infile = fopen(fedlogfile, "r");
 if (infile==0)
  {
   printf("Error: Could not open federal return '%s'\n", fedlogfile);
   fprintf(outfile,"\nError: Could not open federal return '%s'\n", fedlogfile);
   exit(1);
  }
 printf(" Reading file: %s\n", fedlogfile );

 /* Set initial default values. */
 fed_data->Itemized = 1; 
 fed_data->Limited = 1; 
 fed_data->Limited_L6 = 1; 
 for (linenum=0; linenum<MAX_LINES; linenum++) fed_data->fedline[linenum] = 0.0;
 for (linenum=0; linenum<MAX_LINES; linenum++) fed_data->Sched_A[linenum] = 0.0;

 read_line(infile,fline);  linenum = 0;
 while (!feof(infile))
  {
   if ( verbose ) printf( "Read Line: %s" ,fline );
   if (strstr(fline,"Use standard deduction.")!=0) fed_data->Itemized = 0;
   if (strstr(fline,"Deductions not limited")!=0) 
    {
     fed_data->Limited = 0;
     if ( strstr(fline,"line 6") ) fed_data->Limited_L6 = 0;
    }
   next_word(fline, word, " \t=");
   if ((strstr(word,"A")==word) && (isdigit(word[1])) && (strstr(fline," = ")!=0))
    {
     if (sscanf(&word[1],"%d",&linenum)!=1) printf("Error: Reading fed sched-A line number '%s%s'\n",word,fline);
     next_word(fline, word, " 	=");
     if (sscanf(word,"%lf", &fed_data->Sched_A[linenum])!=1) printf("Error: Reading fed sched-A line %d '%s%s'\n",linenum,word,fline);
     if (verbose) printf("Sched_A[%d] = %2.2f\n", linenum, fed_data->Sched_A[linenum]);
    }
   if ((strstr(word,"L")==word) && (strstr(fline," = ")!=0))
    {
     if (sscanf(&word[1],"%d",&linenum)!=1) printf("Error: Reading fed line number '%s%s'\n",word,fline);
     next_word(fline, word, " 	=");
     if (sscanf(word,"%lf", &fed_data->fedline[linenum])!=1) printf("Error: Reading fed line %d '%s%s'\n",linenum,word,fline);
     if (verbose) printf("FedLin[%d] = %2.2f\n", linenum, fed_data->fedline[linenum]);
    }
   read_line(infile,fline);
  }
 fclose(infile);
}


/* ------------------------------------------------------------------------------------ */
/* ---				Main						    --- */
/* ------------------------------------------------------------------------------------ */
int main( int argc, char *argv[] )
{
 int i, j, k, status, Exception, exmp, Exemptions=0;
 char word[1000], outfname[1000];
 time_t now;
 struct FedReturnData fed_data;
 double AB=0.0, ws[MAX_LINES], stdded, min_payment = 0.0, min2file;
 double L23a=0.0, L23b=0.0, PYR33a=0.0, PYR33b=0.0;
 double L24a=0.0, L24b=0.0, L24c=0.0, L24d=0.0;
 double L44a=0.0, L44b=0.0, L44c=0.0, L44d=0.0;
 double L45a=0.0, L45b=0.0;

 /*-----------------------------------------*/
 /* --- Decode any command line options. -- */
 /*-----------------------------------------*/
 printf("NC D400 2012 - v%3.2f\n", thisversion);
 i = 1;  k=1;
 while (i < argc)
 {
  if (strcmp(argv[i],"-verbose")==0)  { verbose = 1; }
  else
  if (k==1)
   {
    infile = fopen(argv[i],"r");
    if (infile==0) {printf("ERROR: Parameter file '%s' could not be opened.\n", argv[i]); exit(1);}
    k = 2;
    /* Base name of output file on input file. */
    strcpy(outfname,argv[i]);
    j = strlen(outfname)-1;
    while ((j>=0) && (outfname[j]!='.')) j--;
    if (j<0) strcat(outfname,"_out.txt"); else strcpy(&(outfname[j]),"_out.txt");
    outfile = fopen(outfname,"w");
    if (outfile==0) {printf("ERROR: Output file '%s' could not be opened.\n", outfname); exit(1);}
    printf("Writing results to file:  %s\n", outfname);
   }
  else
   {printf("Unknown command-line parameter '%s'\n", argv[i]); exit(1);}
  i = i + 1;
 }

 if (infile==0) {printf("Error: No input file on command line.\n"); exit(1);}
 

 /*--------------------------*/
 /* ---- Get Input Data ---- */
 /*--------------------------*/

 /* Pre-initialize all lines to zeros. */
 for (i=0; i<MAX_LINES; i++) { L[i] = 0.0;  ws[i] = 0.0; }

 /* Accept parameters from input file. */
 /* Expect  D-400 lines, something like:
	Title:  NC-400 1999 Return
	L6	34900.0  {Wages}
 */

 /* Accept Form's "Title" line, and put out with date-stamp for your records. */
 read_line( infile, word );
 now = time(0);
 fprintf(outfile,"\n%s		%s\n", word, ctime( &now ));


 /* get_parameter(infile, kind, x, emssg ) */
 get_parameter( infile, 's', word, "FedReturn" );	/* File name of Federal Return log file. */
 if ( verbose ) printf( "word: %s\n", word );
 get_word( infile, word ); 
 ImportFederalReturnData( word, &fed_data);

 get_parameter( infile, 's', word, "Status"); /* 1=single, 2=married/joint, 3=married/separate, 4=house-head, 5=widow */
 get_parameter( infile, 'l', word, "Status ?");
 if ((word[0]>'0') && (word[0]<'6')) status = word[0]-48; else
 if (strncasecmp(word,"Single",4)==0) status = SINGLE; else
 if (strncasecmp(word,"Married/Joint",13)==0) status = MARRIED_FILLING_JOINTLY; else
 if (strncasecmp(word,"Married/Sep",11)==0) status = MARRIED_FILLING_SEPARAT; else
 if (strncasecmp(word,"Head_of_House",4)==0) status = HEAD_OF_HOUSEHOLD; else
 if (strncasecmp(word,"Widow",4)==0) status = WIDOW;
 else
  { 
   printf("Error: unrecognized status '%s'. Must be: Single, Married/joint, Married/sep, Head_of_house, Widow(er)\nExiting.\n", word); 
   fprintf(outfile,"Error: unrecognized status '%s'. Must be: Single, Married/joint, Married/sep, Head_of_house, Widow(er)\nExiting.\n", word); 
   exit(1); 
  }
 fprintf(outfile,"Status = %s (%d)\n", word, status);


 /* Answer yes, if you are claimed on someone else's fed income */
 /* married filing separate and spouse claims deductions, else answer no. */
 get_parameter( infile, 's', word, "Exception" );
 get_parameter( infile, 'b', &Exception, "Exception ?"); 
 if (Exception==1) { printf("Exception case not handled.\n"); fprintf(outfile,"Exception case not handled.\n"); exit(1);} 

 get_parameter( infile, 's', word, "YouAge65" );
 get_parameter( infile, 'b', &Exemptions, "YouAge65 ?"); 
 get_parameter( infile, 's', word, "YouBlind" );
 get_parameter( infile, 'b', &exmp, "YouBlind ?"); 
 Exemptions = Exemptions + exmp;
 get_parameter( infile, 's', word, "SpouseAge65" );
 get_parameter( infile, 'b', &exmp, "SpouseAge65 ?"); 
 Exemptions = Exemptions + exmp;
 get_parameter( infile, 's', word, "SpouseBlind" );
 get_parameter( infile, 'b', &exmp, "SpouseBlind ?"); 
 Exemptions = Exemptions + exmp;

 GetLine( "L19", &L[19] );	/* Tax credits. (D-400TC part 4 line 37) */  
 GetLine( "L21", &L[21] );	/* Consumer Use Tax. (pg 8) */
 GetLine( "L23a", &L23a );	/* Your North Carolina Income Tax Withheld */
 GetLine( "L23b", &L23b );	/* Spouse's North Carolina Income Tax Withheld */
 L[23] = L23a + L23b;
 GetLine( "L24a", &L24a );	/* Other tax payments. 2012 Estimated Tax. */
 GetLine( "L24b", &L24b );	/* Other tax payments. Paid with Extension. */
 GetLine( "L24c", &L24c );	/* Other tax payments. Partnership. */
 GetLine( "L24d", &L24d );	/* Other tax payments. S Corporation. */
 L[24] = L24a + L24b + L24c + L24d;

 GetLine( "L25", &L[25] );	/* North Carolina Earned Income Tax Credit (D-400TC, Part 5) */

 GetLine( "L34", &L[34] );	/* Interest income from non-NC states. */
 GetLine( "L35", &L[35] );	/* Adj Bonus depreciation */
 GetLine( "L36", &L[36] );	/* Other additions to fed AGI */
 L[37] = L[34] + L[35] + L[36];
 L[7] = L[37];

 GetLine( "L38", &L[38] );	/* State or local income tax refund not in Fed 1040 line 10 */
 GetLine( "L39", &L[39] );	/* Interest from US or US possessions. */
 GetLine( "L40", &L[40] );	/* Taxable soc sec and RR benfits. */
 GetLine( "L41", &L[41] );	/* Retirement benefits (Bailey)*/
 GetLine( "L42", &L[42] );	/* Other retirement benefits */
 GetLine( "L43", &L[43] );	/* Severance wages (See Line instructions on Page 13) */
 GetLine( "L44a", &L44a );	/* Adjustmnts for bonus depreciation added back 2008 */
 GetLine( "L44b", &L44b );	/* Adjustmnts for bonus depreciation added back 2009 */
 GetLine( "L44c", &L44c );	/* Adjustmnts for bonus depreciation added back 2010 */
 GetLine( "L44d", &L44d );	/* Adjustmnts for bonus depreciation added back 2011 */
 L[44] = L44a + L44b + L44c + L44d;
 GetLine( "L45a", &L45a );	/* Adjustment for Section 179 Expense Added Back in 2010 */
 GetLine( "L45b", &L45b );	/* Adjustment for Section 179 Expense Added Back in 2011 */
 L[45] = L45a + L45b;

 GetLine( "L46", &L[46] );	/* Contributions to North Carolina's National College Savings Program */
 GetLine( "L47", &L[47] );	/* Adjustment for absorbed Non-ESB NOL added back in 2003 - 2006  */
 GetLine( "L48", &L[48] );	/* Adj for net business income not passive */
 GetLine( "L49", &L[49] );	/* Other deductions from fed income */

 GetLine( "PYR-ws33a", &PYR33a ); /* Part-year resident worksheet line 33a. */
 GetLine( "PYR-ws33b", &PYR33b ); /* Part-year resident worksheet line 33b. */
 if (PYR33a == 0.0)
  L[16] = 1.0;
 else
  {
   L[54] = PYR33b;
   L[55] = PYR33a;
   L[56] = 0.0001 * (int)(10000.0 * (L[54] / L[55]) + 0.5);	/* Round to four decimal places. */
   L[16] = L[56];
  }

 fclose(infile);


 /*-------------------------------*/
 /* ---- Do Tax Calculations ---- */
 /*-------------------------------*/

 L[6] = fed_data.fedline[37];		/* Taxable income from Fed1040 Line 37. */

 /* Exemptions calculation of Deductions - pg 11 */
 switch (status)
  {
   case SINGLE: 		 stdded = 3000.0; 	/* NC std single deduction. */
				 min2file = 5500.0;
	break;
   case MARRIED_FILLING_JOINTLY: stdded = 6000.0; 	/* NC std Married/joint deduction. */
				 min2file = 11000.0;
	break;
   case WIDOW:			 stdded = 6000.0; 	/* NC std widow(er) deduction. */
				 min2file = 8500.0;
	break;
   case MARRIED_FILLING_SEPARAT: stdded = 3000.0; 	/* NC std Married/sep deduction. */
				 min2file = 2500.0;
	break;
   case HEAD_OF_HOUSEHOLD:	 stdded = 4400.0; 	/* NC std Head of house deduction. */
				 min2file = 6900.0;
	break;
   default:  stdded = 0;  printf("Unknown status\n");  exit(1); 
  }

 if (L[6] <= min2file)
  fprintf(outfile, "You may not need to file NC tax return, due to your income.\n");

 if ( Exemptions > 0 )
  {
     switch (status)
      {
       case SINGLE: /* NC single deduction. */
		if (Exemptions == 1) stdded = 3750.0;
		else stdded = 4500.0;
        break;
       case MARRIED_FILLING_JOINTLY: /* NC Married/joint deduction. */
       case WIDOW: /* NC widow(er) deduction. */
		if (Exemptions == 1) stdded = 6600.0;
		if (Exemptions == 1) stdded = 7200.0;
		if (Exemptions == 1) stdded = 7800.0;
		else stdded = 8400.0;
        break;
       case MARRIED_FILLING_SEPARAT: /* NC Married/sep deduction. */
		if (Exemptions == 1) stdded = 3600.0;
		if (Exemptions == 1) stdded = 4200.0;
		if (Exemptions == 1) stdded = 4800.0;
		else stdded = 5400.0;
        break;
       case HEAD_OF_HOUSEHOLD: /* NC Head of house deduction. */
		if (Exemptions == 1) stdded = 5150.0;
		else stdded = 5900.0;
        break;
      }   
  }

 L[8] = L[6] + L[7];

 for (j=38; j <= 59; j++)
  L[50] = L[50] + L[j];

 L[9] = L[50];			/* Deductions from federal taxable income. */	

 L[10] = L[8] - L[9];

 if (fed_data.Itemized) 
  {
   double it_ded_ws[20];
   it_ded_ws[1] = fed_data.Sched_A[29];
   it_ded_ws[2] = stdded;
   it_ded_ws[3] = NotLessThanZero( it_ded_ws[1] - it_ded_ws[2] );
   it_ded_ws[4] = fed_data.Sched_A[5] + fed_data.Sched_A[8];
   it_ded_ws[5] = smallerof( it_ded_ws[3], it_ded_ws[4] );
   it_ded_ws[6] = NotLessThanZero( it_ded_ws[1] - it_ded_ws[5] );
   L[11] = it_ded_ws[6];
  }
 else
  L[11] = stdded;

 L[12] = L[10] - L[11];

 /* Personal Exemption Adjustment worksheet - pg 9 */
 switch (status)
  {
   case SINGLE: 		 AB =  60000.0; break;  /* NC std single */
   case MARRIED_FILLING_JOINTLY: AB = 100000.0; break;  /* NC std Married/joint */
   case WIDOW:			 AB = 100000.0; break;  /* NC std widow(er) */
   case MARRIED_FILLING_SEPARAT: AB =  50000.0; break;  /* NC std Married/sep */
   case HEAD_OF_HOUSEHOLD:	 AB =  80000.0; break;  /* NC std Head of house */
  }
 if (fed_data.fedline[37] < AB)		/* Compare Fed AGI */
    L[13] = fed_data.fedline[6] * 2500.0;	/* A pg 9 */
 else
    L[13] = fed_data.fedline[6] * 2000.0;	/* B pg 9 */

 L[14] = L[12] - L[13];

 L[15] = L[14];

 L[17] = L[15];		/* North Carolina Taxable Income. */
				/*  (Assumes full-year resident of NC.) */

 if (L[17] <= 68000.0)
  {
   int m;
   double x;
   m = L[17] / 50.0;    /* Round income to nearest $50. */
   x = TaxRateFunction( 50.0 * (double)m + 25.0, status );       /* Lookup. */
   m = (int)(x + 0.5);  /* Round result to whole dollar. */
   L[18] = (double)m;
  }
 else L[18] = TaxRateFunction( L[17], status );

 L[20] = L[18] - L[19];

/* Calculate USE tax if not entered on L21
 * based on Use Tax Worksheet on page 10.
 *
 * If you have purchases greater that 1000 that need
 * to be reported, or taxes paid to another state, then
 * fill out the Work Sheet in the instructions and enter
 * the amount on L21 in data file.
 */

 printf( "Assuming you have calculated your USE tax (%2.2f) according to instructions pg 8\n", L[17] );

 L[22] = L[20] + L[21];

 L[26] = L23a + L23b + L[24] + L[25];

 if (L[22] > L[26]) 
  {
   L[27] = L[22] - L[26];	/* You OWE */

   /* Check for under payment see form D422 Part I */
   min_payment = L[22] * 0.9; /* min est tax payments see form D422 Part I */
   if ((L[26] < min_payment) && (L[27] > 1000.00)) 
    {
     /* We would calc penalty here... */
     printf("WARNING: Possible underpayment of est. taxes penalty. Calculation not performed.\n"); 
    }
   L[28] = L[27];  /* Assumes no penalties. */
  }
 else
  {
   L[29] = L[26] - L[22];	/* REFUND */
  }



 /*-------------------------*/
 /* ---- Print Results ---- */
 /*-------------------------*/

 showline(6);	/* Taxable fed income */
 showline(7);	/* Additions to fed income */
 showline(8);	
 showline(9);	/* Deductions */
 showline(10);
 showline(11);
 showline(12);	
 showline(13);	
 showline(14);	
 showline(15);	
 if (L[16] < 1.0) showline(16);  /* Part-yr */	
 showline_wmsg(17,"North Carolina Taxable Income");
 showline_wmsg(18,"North Carolina Income Tax");
 showline(19);
 showline(20);
 showline(21);
 showline(22);
 showline_wlabelmsg( "L23a", L23a, "Your North Carolina Tax Withheld");
 showline_wlabelmsg( "L23b", L23b, "Spouse's North Carolina Tax Withheld");
 showline_wlabel( "L24a", L24a );
 showline_wlabel( "L24b", L24b );
 showline_wlabel( "L24c", L24c );
 showline_wlabel( "L24d", L24d );
 showline(21);
 showline(22);
 showline(23);
 if (L[22] > L[26])
  {
   showline(27);
   showline_wmsg(28,"PAY THIS AMOUNT");
   if ((L[26] < min_payment) && (L[27] > 1000.00))
    {
     fprintf(outfile," You may owe underpayment interest and penalties.\n");
     fprintf(outfile," See page 6 instructions to calculate them according to your situation.\n");
    }
  }
 else
  {
   showline_wmsg(29,"OVERPAYMENT");
  }

 for (k=34; k<=43; k++)
  showline(k);

 showline_wlabel( "L44a", L44a );
 showline_wlabel( "L44b", L44b );
 showline_wlabel( "L44c", L44c );
 showline_wlabel( "L44d", L44d );
 showline_wlabel( "L44e", L[44] );

 showline_wlabel( "L45a", L45a );
 showline_wlabel( "L45b", L45b );
 showline_wlabel( "L45c", L[45] );

 for (k=46; k<=50; k++)
  showline(k);

 if (PYR33a > 0.0)
  for (k=51; k<=53; k++)
   showline(k);

 fclose(outfile);
 Display_File( outfname );
 printf("\nResults written to file '%s'\n", outfname);
 return 0;
}
